<?php
/*
Copyright 2010 Main Street Softworks, Inc. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

   1. Redistributions of source code must retain the above copyright notice, this list of
      conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright notice, this list
      of conditions and the following disclaimer in the documentation and/or other materials
      provided with the distribution.

THIS SOFTWARE IS PROVIDED BY MAIN STREET SOFTWORKS INC ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL MAIN STREET SOFTWORKS INC OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

The views and conclusions contained in the software and documentation are those of the
authors and should not be interpreted as representing official policies, either expressed
or implied, of Main Street Softworks, Inc.
*/



/* Lets make it so this may always be included, even if
 * they have the module loaded for php_mcve.  Prefer the
 * module routines over the pure-php ones */
if (!function_exists("m_initengine")) {

//error_reporting(E_ALL);

define('M_CONN_SSL', 1);
define('M_CONN_IP',  2);

define('M_TRAN_STATUS_NEW', 1);
define('M_TRAN_STATUS_SENT', 2);
define('M_TRAN_STATUS_DONE', 3);

define('M_ERROR', -1);
define('M_FAIL', 0);
define('M_SUCCESS', 1);

define('M_DONE', 2);
define('M_PENDING', 1);

$init_cafile = "";
function M_InitEngine($cafile)
{
	global $init_cafile;
	$init_cafile = $cafile;
}

function M_DestroyEngine()
{
	/* Do nothing */
}

function M_InitConn()
{
	global $init_cafile;
	$conn = array();
	$conn['blocking']        = false;
	$conn['conn_error']      = "";
	$conn['conn_timeout']    = 10;
	$conn['host']            = "";
	$conn['last_id']         = 0;
	$conn['method']          = M_CONN_IP;
	$conn['port']            = 0;
	$conn['readbuf']	 = null;
	$conn['ssl_cafile']      = $init_cafile;
	$conn['ssl_verify']      = false;
	$conn['ssl_cert']        = null;
	$conn['timeout']         = 0;
	$conn['tran_array']      = array();
	$conn['verify_conn']     = true;
	$conn['writebuf']        = null;
	return $conn;
}

function M_SetIP(&$conn, $host, $port)
{
	$conn['host']   = $host;
	$conn['port']   = $port;
	$conn['method'] = M_CONN_IP;	
	return true;
}

function M_SetSSL(&$conn, $host, $port)
{
	$conn['host']   = $host;
	$conn['port']   = $port;
	$conn['method'] = M_CONN_SSL;
	return true;
}

function M_SetDropFile(&$conn, $directory)
{
	/* NOT SUPPORTED! */
	return false;
}

function M_verifyping(&$conn)
{
	$blocking = $conn['blocking'];
	M_SetBlocking($conn, false);
	$id = M_TransNew($conn);
	M_TransKeyVal($conn, $id, "action", "ping");
	if (!M_TransSend($conn, $id)) {
		M_DeleteTrans($conn, $id);
		return false;
	}
	$lasttime=time();
	while (M_CheckStatus($conn, $id) == M_PENDING && time()-$lasttime <= 5) {
		if (!M_Monitor($conn))
			break;
		M_uwait(100000);
	}
	M_SetBlocking($conn, $blocking);
	$status = M_CheckStatus($conn, $id);
	M_DeleteTrans($conn, $id);
	if ($status != M_DONE)
		return false;
	return true;
}

function M_Connect(&$conn)
{
	$ctx = stream_context_create();

	if ($conn['method'] == M_CONN_SSL) {
		stream_context_set_option($ctx, 'ssl', 'cafile', $conn['ssl_cafile']);
		stream_context_set_option($ctx, 'ssl', 'verify_peer', $conn['ssl_verify']);
		stream_context_set_option($ctx, 'ssl', 'allow_self_signed', !$conn['ssl_verify']);
		if ($conn['ssl_cert'] != NULL) {
			stream_context_set_option($ctx, 'ssl', 'local_cert', $conn['ssl_cert']);
		}
	}
	
	if ($conn['method'] == M_CONN_SSL)
		$url = "ssl://" . $conn['host'] . ":" . $conn['port'];
	else
		$url = "tcp://" . $conn['host'] . ":" . $conn['port'];

	$error = "";
	$errorString = "";
	$conn['fd'] = stream_socket_client($url, $error, $errorString, $conn['conn_timeout'], STREAM_CLIENT_CONNECT, $ctx);
	if (!$conn['fd']) {
		$conn['conn_error'] = "Failed to connect to $url: $error: $errorString";
		return false;
	}

	if ($conn['verify_conn'] && !M_verifyping($conn)) {
		$conn['conn_error'] = "PING request failed";
		fclose($conn['fd']);
		$conn['fd'] = false;
		return false;
	}
	return true;
}

function M_ConnectionError(&$conn)
{
	return $conn['conn_error'];
}

function M_DestroyConn(&$conn)
{
	fclose($conn['fd']);
	$conn['fd'] = false;
	unset($conn);
}

function M_MaxConnTimeout(&$conn, $secs)
{
	$conn['conn_timeout'] = $secs;
	return true;
}

function M_SetBlocking(&$conn, $tf)
{
	if ($tf)
		$conn['blocking'] = true;
	else
		$conn['blocking'] = false;

	return true;
}

function M_ValidateIdentifier(&$conn, $tf)
{
	/* Always validated, stub for compatibility */
	return true;
}

function M_VerifyConnection(&$conn, $tf)
{
	if ($tf)
		$conn['verify_conn'] = true;
	else
		$conn['verify_conn'] = false;

	return true;
}

function M_VerifySSLCert(&$conn, $tf)
{
	if ($tf)
		$conn['ssl_verify'] = true;
	else
		$conn['ssl_verify'] = false;

	return true;
}

function M_SetSSL_CAfile(&$conn, $cafile)
{
	$conn['cafile'] = $cafile;
	return true;
}

function M_SetSSL_Files(&$conn, $sslkeyfile, $sslcertfile)
{
	/* PHP doesn't allow separate key and cert files */
	if ($sslkeyfile != NULL && $sslcertfile != NULL &&
	    strcmp($sslkeyfile, $sslcertfile) != 0)
		return false;
	if ($sslkeyfile == NULL && $sslcertfile == NULL)
		return false;

	if ($sslkeyfile != NULL)
		$conn['ssl_cert'] = $sslkeyfile;
	else
		$conn['ssl_cert'] = $sslcertfile;
	return true;
}

function M_SetTimeout(&$conn, $secs)
{
	$conn['timeout'] = $secs;
	return true;
}

function M_TransNew(&$conn)
{
	$tran = array();
	$tran['id']              = ++$conn['last_id'];
	$tran['status']          = M_TRAN_STATUS_NEW;
	$tran['comma_delimited'] = false;
	$tran['in_params']       = array();
	$tran['out_params']      = array();
	$tran['raw_response']    = null;
	$tran['csv']             = null;
	$conn['tran_array'][$tran['id']] = &$tran;
	return $tran['id'];
}

function &M_findtranbyid(&$conn, $id)
{
	if (!isset($conn['tran_array'][$id]))
		return false;

	return $conn['tran_array'][$id];
}

function M_TransKeyVal(&$conn, $id, $key, $val)
{
	$tran =& M_findtranbyid($conn, $id);
	/* Invalid ptr, or transaction has already been sent out */
	if ($tran === false || $tran['status'] != M_TRAN_STATUS_NEW)
		return false;
	
	$tran['in_params'][$key] = $val;
	
	return true;
}

function M_TransBinaryKeyVal(&$conn, $id, $key, $val, $val_len)
{
	return M_TransKeyVal($conn, $id, $key, base64_encode($val));
}

function M_TransSend(&$conn, $id)
{
	$tran =& M_findtranbyid($conn, $id);
	/* Invalid ptr, or transaction has already been sent out */
	if ($tran === false || $tran['status'] != M_TRAN_STATUS_NEW)
		return false;

	$tran['status'] = M_TRAN_STATUS_SENT;
	/* Structure Transaction */
	
	/* STX, identifier, FS */
	$tran_str = "\x02" . $tran['id'] . "\x1c";

	/* PING is specially formed */
	if (isset($tran['in_params']['action']) &&
	    strcasecmp($tran['in_params']['action'], "ping") == 0) {
		$tran_str .= "PING";
	} else {
		/* Each key/value pair in array as key="value" */
		foreach ($tran['in_params'] as $key => $value) {
			$tran_str .= $key . '="' . str_replace('"', '""', $value) . '"' . "\r\n"; 
		}
		/* Add timeout if necessary */
		if ($conn['timeout'] != 0) {
			$tran_str .= 'timeout="' . $conn['timeout'] . '"' . "\r\n";
		}
	}
	
	/* ETX */
	$tran_str .= "\x03";

	$conn['writebuf'] .= $tran_str;

	if ($conn['blocking']) {
		while (M_CheckStatus($conn, $id) == M_PENDING) {
			if (!M_Monitor($conn))
				return false;
			M_uwait(100000);
		}
	}
	return true;
}

function M_CheckStatus(&$conn, $id)
{
	$tran =& M_findtranbyid($conn, $id);
	if ($tran === false ||
            ($tran['status'] != M_TRAN_STATUS_SENT && $tran['status'] != M_TRAN_STATUS_DONE))
		return M_ERROR;

	if ($tran['status'] == M_TRAN_STATUS_SENT)
		return M_PENDING;

	return M_DONE;
}

function M_CompleteAuthorizations(&$conn, &$id_array)
{
	$id_array = array();
	for ($i=0; $i<count($conn['tran_array']); $i++) {
		if ($conn['tran_array'][$i]['status'] == M_TRAN_STATUS_DONE)
			$id_array[] = $conn['tran_array'][$i]['id'];
	}
	return count($id_array);
}

function M_DeleteTrans(&$conn, $id)
{
	$tran =& M_findtranbyid($conn, $id);
	if ($tran === false)
		return false;
	unset($conn['tran_array'][$id]);
	return true;
}

/* Return false on error, 0 if no bytes pending, 1 on bytes pending */
function M_BytesPending($fd, $to_usec)
{
	$readarr[] = $fd;
	$writearr  = null;
	$exceptarr = null;
	$tv_sec    = 0;
	$tv_usec   = $to_usec;
	
	return stream_select($readarr, $writearr, $exceptarr, $tv_sec, $tv_usec);

}

function M_verify_comma_delimited($data)
{
	for ($i=0; $i<strlen($data); $i++) {
		/* If hit a new line or a comma before an equal sign, must
		 * be comma delimited */
		if (ord($data[$i]) == 0x0A ||
		    ord($data[$i]) == 0x0D ||
		    $data[$i] == ',')
			return true;
		/* If hit an equal sign before a new line or a comma, must be
		 * key/val */
		if ($data[$i] == '=')
			return false;
	}
	/* Who knows?  Should never get here */
	return true;
}

function M_explode_quoted($delim, $data, $quote_char)
{
	$myarr = array();
	$on_quote = 0;
	$beginsect = 0;
	for ($i=0; $i<strlen($data); $i++) {
		if ($data[$i] == $quote_char) {
			/* Doubling the quote char acts as escaping */
			if ($data[$i+1] == $quote_char) {
				$i++;
				continue;
			} else if ($on_quote) {
				$on_quote = 0;
			} else {
				$on_quote = 1;
			}
		}
		if ($data[$i] == $delim && !$on_quote) {
			$myarr[] = substr($data, $beginsect, $i-$beginsect);
			$beginsect = $i+1;
		}
	}
	if (count($myarr) == 0 && strlen($data)) {
		$myarr[] = $data;
	}
	return $myarr;
}

function M_remove_dupe_quotes($data)
{
	if (strpos($data, '"') === false)
		return $data;

	$ret = "";
	for ($i=0; $i<strlen($data); $i++) {
		if ($data[$i] == '"' && $data[$i+1] == '"') {
			$ret .= '"';
			$i++;
		} else if ($data[$i] != '"') {
			$ret .= $data[$i];
		}
	}
	return $ret;
}

function M_Monitor(&$conn)
{
	if (!$conn['fd'])
		return false;

	/* Read Data */
	do {
		$data_read = false;

		$retval = M_BytesPending($conn['fd'], 20000);
		if ($retval === false) {
			fclose($conn['fd']);
			$conn['fd'] = false;
			$conn['conn_error'] = "select failure";
			return false;
		} else if ($retval > 0) {
			$buf = fread($conn['fd'], 8192);
			if ($buf === false) {
				fclose($conn['fd']);
				$conn['fd'] = false;
				$conn['conn_error'] = "fread failure";
				return false;
			}
			$conn['readbuf'] .= $buf;
			$data_read = true;
		}
	} while($data_read);

	/* Write Data */
	if (strlen($conn['writebuf'])) {
		$retval = fwrite($conn['fd'], $conn['writebuf'], strlen($conn['writebuf']));
		if ($retval === false) {
			fclose($conn['fd']);
			$conn['fd'] = false;
			$conn['conn_error'] = "fwrite failure";
			return false;
		} else if ($retval == strlen($conn['writebuf'])) {
			$conn['writebuf'] = null;
		} else {
			$oldlen = strlen($conn['writebuf']);
			$conn['writebuf'] = substr($conn['writebuf'], $retval, $oldlen-$retval);
		}
	}

	/* Parse */
	while(strlen($conn['readbuf'])) {
		if ($conn['readbuf'][0] != chr(0x02)) {
			fclose($conn['fd']);
			$conn['fd'] = false;
			$conn['conn_error'] = "protocol error, responses must start with STX";
			return false;
		}
		$etx = strpos($conn['readbuf'], chr(0x03));
		if ($etx === false) {
			/* Not enough data */
			break;
		}

		/* Chop off txn from readbuf */
		$readbuf_len = strlen($conn['readbuf']);
		$txndata = substr($conn['readbuf'], 0, $etx);
		if ($etx+1 == $readbuf_len) {
			$conn['readbuf'] = null;
		} else {
			$temp = substr($conn['readbuf'], $etx+1, $readbuf_len-($etx+1));
			$conn['readbuf'] = $temp;
		}

		$fs = strpos($txndata, chr(0x1c));
		if ($fs === false) {
			fclose($conn['fd']);
			$conn['fd'] = false;
			$conn['conn_error'] = "protocol error, responses must contain a FS";
			return false;
		}

		$id = substr($txndata, 1, $fs - 1);
		$data = substr($txndata, $fs+1, strlen($txndata)-$fs-1);

		$txn = &M_findtranbyid($conn, intval($id));
		if ($txn === false) {
			echo "Unrecognized identifier in response: '$id'\n";
			/* Discarding data */
			continue;
		}

		$txn['status']          = M_TRAN_STATUS_DONE;
		$txn['raw_response']    = $data;
		$txn['comma_delimited'] = M_verify_comma_delimited($data);
		if (!$txn['comma_delimited']) {
			$lines = M_explode_quoted("\n", $data, '"');
			for ($i=0; $i<count($lines); $i++) {
				$lines[$i] = trim($lines[$i]);
				if (strlen($lines[$i])) {
					$keyval = explode("=", $lines[$i], 2);
					$txn['out_params'][$keyval[0]] = M_remove_dupe_quotes(trim($keyval[1]));
				}
			}
		}
	}
	return true;
}

function M_TransInQueue(&$conn)
{
	return count($conn['tran_array']);
}

function M_TransactionsSent(&$conn)
{
	if (strlen($conn['writebuf']))
		return false;
	return true;
}

function M_GetCell(&$conn, $id, $col, $row)
{
	$tran =& M_findtranbyid($conn, $id);
	if ($tran === false)
		return false;

	/* Invalid ptr, or transaction has not returned */
	if ($tran['status'] != M_TRAN_STATUS_DONE)
		return false;
	if (!$tran['comma_delimited'])
		return false;
	if (!isset($tran['csv'][$row+1]))
		return false;
	for ($i=0; $i<count($tran['csv'][0]); $i++) {
		if (strcasecmp($tran['csv'][0][$i], $col) == 0)
			return $tran['csv'][$row+1][$i];
	}
	return false;
}

function M_GetBinaryCell(&$conn, $id, $col, $row, &$outlen)
{
	$outlen = 0;
	$out = null;
	$cell = M_GetCell($conn, $id, $col, $row);
	if ($cell)
		$out = base64_decode($cell);
	if ($out)
		$outlen = strlen($out);
	return $out;
}

function M_GetCellByNum(&$conn, $id, $col, $row)
{
	$tran =& M_findtranbyid($conn, $id);
	if ($tran === false)
		return false;

	/* Invalid ptr, or transaction has not returned */
	if ($tran['status'] != M_TRAN_STATUS_DONE)
		return false;
	if (!$tran['comma_delimited'])
		return false;
	if (!isset($tran['csv'][$row+1][$col]))
		return false;
	return $tran['csv'][$row+1][$col];
}

function M_GetCommaDelimited(&$conn, $id)
{
	$tran =& M_findtranbyid($conn, $id);
	if ($tran === false)
		return false;

	/* Invalid ptr, or transaction has not returned */
	if ($tran['status'] != M_TRAN_STATUS_DONE)
		return false;
	return $tran['raw_response'];
}

function M_GetHeader(&$conn, $id, $col)
{
	$tran =& M_findtranbyid($conn, $id);
	if ($tran === false)
		return false;

	/* Invalid ptr, or transaction has not returned */
	if ($tran['status'] != M_TRAN_STATUS_DONE)
		return false;
	if (!$tran['comma_delimited'])
		return false;
	if (!isset($tran['csv'][0][$col]))
		return false;
	return $tran['csv'][0][$col];
}

function M_IsCommaDelimited(&$conn, $id)
{
	$tran =& M_findtranbyid($conn, $id);
	if ($tran === false)
		return false;

	/* Invalid ptr, or transaction has not returned */
	if ($tran['status'] != M_TRAN_STATUS_DONE)
		return false;
	return $tran['comma_delimited'];
}

function M_NumColumns(&$conn, $id)
{
	$tran =& M_findtranbyid($conn, $id);
	if ($tran === false)
		return false;

	/* Invalid ptr, or transaction has not returned */
	if ($tran['status'] != M_TRAN_STATUS_DONE)
		return false;
	if (!$tran['comma_delimited'])
		return false;
	return count($tran['csv'][0]);
}

function M_NumRows(&$conn, $id)
{
	$tran =& M_findtranbyid($conn, $id);
	if ($tran === false)
		return false;

	/* Invalid ptr, or transaction has not returned */
	if ($tran['status'] != M_TRAN_STATUS_DONE)
		return false;
	if (!$tran['comma_delimited'])
		return false;
	return count($tran['csv'])-1;
}

/**
* Create a 2D array from a CSV string
*
* @param mixed $data 2D array
* @param string $delimiter Field delimiter
* @param string $enclosure Field enclosure
* @param string $newline Line seperator
* @return
*/
function M_parsecsv($data, $delimiter = ',', $enclosure = '"', $escape = '"')
{
	$pos = $last_pos = -1;
	$end = strlen($data);
	$row = 0;
	$quote_open = false;
	$trim_quote = false;
	$newline = "\n";
	$return = array();

	/* Create a continuous loop */
	while(1) {
		++$pos;
		/* Get the positions */
		$comma_pos = strpos($data, $delimiter, $pos);
		$quote_pos = strpos($data, $enclosure, $pos);
		$newline_pos = strpos($data, $newline, $pos);

		/* Which one comes first? */
		$pos = min(($comma_pos === false) ? 
		$end : $comma_pos, ($quote_pos === false) ? $end :
		$quote_pos, ($newline_pos === false) ? $end : $newline_pos);

		/* Cache it */
		$char = (isset($data[$pos])) ? $data[$pos] : null;
		$done = ($pos == $end);

		/* It it a special character? */
		if ($done || $char == $delimiter || $char == $newline){
			/* Ignore it as we're still in a quote */
			if ($quote_open && !$done)
				continue;

			$length = $pos - ++$last_pos;

			/* drop newlines */
			if ($char == $newline)
				--$length;

			/* drop RFC4180 quotes */
			if ($trim_quote)
				--$length;
				
			/* And we're done */
			if ($done)
				break;

			/* Get all the contents of this column */
			$return[$row][] = ($length > 0) ?
				str_replace($enclosure . $enclosure, $enclosure,
					substr($data, $last_pos, $length)) : '';

			/* Save the last position */
			$last_pos = $pos;

			/* Next row? */
			if ($char == $newline)
				++$row;

			/* reset RFC4180 quote */
			$trim_quote = false;
		}
		/* Opening/Closing quote for RFC4180 */
		else if ($char == $enclosure) {

			/* It's an opening quote */
			if ($quote_open == false) {
				$quote_open = true;
				$trim_quote = true;

				/* Trim this opening quote? */
				if ($last_pos + 1 == $pos)
					++$last_pos;
			}
			/* It's a closing quote */
			else
				$quote_open = false;
		}
	}
	return $return;
}

function M_ParseCommaDelimited(&$conn, $id)
{
	$tran =& M_findtranbyid($conn, $id);
	if ($tran === false)
		return false;

	/* Invalid ptr, or transaction has not returned */
	if ($tran['status'] != M_TRAN_STATUS_DONE)
		return false;

	$tran['csv'] = M_parsecsv($tran['raw_response'], ",", '"', '"');
	return true;
}

function M_ResponseKeys(&$conn, $id)
{
	$tran =& M_findtranbyid($conn, $id);
	if ($tran === false)
		return false;

	/* Invalid ptr, or transaction has not returned */
	if ($tran['status'] != M_TRAN_STATUS_DONE)
		return false;

	$ret = array();
	foreach ($tran['out_params'] as $key => $value)
		$ret[] = $key;

	return $ret;
}

function M_ResponseParam(&$conn, $id, $key)
{
	$tran =& M_findtranbyid($conn, $id);
	if ($tran === false)
		return false;

	/* Invalid ptr, or transaction has not returned */
	if ($tran['status'] != M_TRAN_STATUS_DONE)
		return false;

	return $tran['out_params'][$key];
}

function M_ReturnStatus(&$conn, $id)
{
	$tran =& M_findtranbyid($conn, $id);
	if ($tran === false)
		return false;

	/* Invalid ptr, or transaction has not returned */
	if ($tran['status'] != M_TRAN_STATUS_DONE)
		return M_ERROR;

	if ($tran['comma_delimited'] == true)
		return M_SUCCESS;

	$code = M_ResponseParam($conn, $id, "code");
	if (strcasecmp($code, "AUTH") == 0 || strcasecmp($code, "SUCCESS") == 0)
		return M_SUCCESS;

	return M_FAIL;
}

function M_uwait($usec)
{
	usleep($usec);
}

} /* function_exists("m_initengine") */
?>
